@@ -2,6 +2,7 @@ import re
 import fsspec
 import tempfile
 import warnings
+import logging
 
 import pandas as pd
 
@@ -116,14 +117,14 @@ class BaseObjectStorageDataset():
             assert type(batch_size) is int and batch_size > 0 and type(shard_size) is int and shard_size > 0, "If batch_size and shard_size are specified, both must be greater than zero integers"
             #TODO: For now default to the smaller size
             if batch_size <= shard_size:
-                warnings.warn(f"defaulting to shard_size of {batch_size}")
+                logging.warn(f"defaulting to shard_size of {batch_size}")
                 shard_size = batch_size
             else:
-                warnings.warn(f"defaulting to batch_size of {shard_size}")
+                logging.warn(f"defaulting to batch_size of {shard_size}")
                 batch_size = shard_size
         elif shard_size is not None and batch_size is None:
             assert type(shard_size) is int and shard_size > 0, "The value for shard_size must be a greater than zero integer"
-            warnings.warn(f"defaulting to batch_size of {shard_size}")
+            logging.warn(f"defaulting to batch_size of {shard_size}")
             batch_size = shard_size
         elif batch_size is not None and shard_size is None:
             assert type(batch_size) is int and batch_size > 0, "The value for batch_size must be a greater than zero integer"
@@ -166,7 +167,10 @@ class BaseObjectStorageDataset():
             _, uri = fsspec.core.split_protocol(glob)
             self.objs = [uri]
         else:
-            self.objs = self.fs.glob(glob)
+            if self.fs.isdir(glob):
+                self.objs = self.fs.glob(f"{remove_trailing_forward_slash(glob)}/*")
+            else:
+                self.objs = self.fs.glob(glob)
 
         if not isinstance(self.objs, list) or not len(self.objs):
             raise RuntimeWarning(f"Specified glob pattern {self.glob} failed to match any objects. Ensure that the permission setting are configured correctly using storage_options before trying again.")
@@ -179,15 +183,29 @@ class BaseObjectStorageDataset():
         else:
             #if not specified explicitly try to fall back to default location
             if type(stats_glob) is not str and protocol not in ('http', 'https'):
-                maybe_stats_glob = f"{protocol}://{self.fs._parent(glob)}/.meta/stats/"
+
+                #if the glob is a file or a wildcard
+                if self.fs.isfile(glob) or (not self.fs.isfile(glob) and not self.fs.isdir(glob)):
+                    maybe_stats_glob = f"{protocol}://{self.fs._parent(glob)}/.meta/stats/"
+                elif self.fs.isdir(glob):
+                    #remove trailing / characters
+                    maybe_stats_glob = f"{remove_trailing_forward_slash(glob)}/.meta/stats/"                
+                else:
+                    logging.warn(f"failed to construct a stats dataframe path for glob={glob}")
+                    maybe_stats_glob = None
+
                 maybe_stats = self.fs.glob(maybe_stats_glob)
                 if not isinstance(maybe_stats, list) or not len(maybe_stats):
-                    warnings.warn(f"stats_glob is not explicitly specified at initialization, failed to find stats metadata at default location {maybe_stats_glob}")
+                    logging.warn(f"stats_glob is not explicitly specified at initialization, failed to find stats metadata at default location {maybe_stats_glob}")
                     stats_glob = None
                 else:
                     [stats_glob] = maybe_stats
                     stats_glob = f"{protocol}://{stats_glob}"
-                    warnings.warn(f"stats_glob is not specified at initialization, defaulting to stats_glob={stats_glob}")
+                    logging.warn(f"stats_glob is not specified at initialization, defaulting to stats_glob={stats_glob}")            
+
+            else:
+                stats_glob = None
+
             
             if stats_glob is not None:
                 self.stats_df = pd.read_csv(f"filecache::{stats_glob}")
@@ -197,19 +215,28 @@ class BaseObjectStorageDataset():
             
             #ignore fall back with protocols that dont support glob
             if shards_glob is None and protocol not in ('http', 'https'):
-                maybe_shards_glob = f"{protocol}://{self.fs._parent(glob)}/.meta/shards/"
+
+                #if the glob is a file or a wildcard
+                if self.fs.isfile(glob) or (not self.fs.isfile(glob) and not self.fs.isdir(glob)):
+                    maybe_shards_glob = f"{protocol}://{self.fs._parent(glob)}/.meta/shards/"
+                elif self.fs.isdir(glob):
+                    #remove trailing / characters
+                    maybe_shards_glob = f"{remove_trailing_forward_slash(glob)}/.meta/shards/"                
+                else:
+                    logging.warn(f"failed to construct a shards dataframe path for glob={glob}")
+                    maybe_shards_glob = None
+                
                 maybe_shards = self.fs.glob(maybe_shards_glob)
                 if not isinstance(maybe_shards, list) or not len(maybe_shards):
-                    warnings.warn(f"shards_glob is not specified at initialization, failed to find shard metadata at default location {maybe_shards_glob}")
+                    logging.warn(f"shards_glob is not specified at initialization, failed to find shard metadata at default location {maybe_shards_glob}")
                 else:
                     [shards_glob] = maybe_shards
                     shards_glob = f"{protocol}://{shards_glob}"
-                    warnings.warn(f"shards_glob is not specified at initialization, defaulting to shards_glob={shards_glob}")
-
+                    logging.warn(f"shards_glob is not specified at initialization, defaulting to shards_glob={shards_glob}")
 
             if shards_glob is None:
                 index_shards = True
-                warnings.warn("shards_glob is not specified at initialization, attempting to proceed with index_shards=True which can take a while for large objects.")
+                logging.warn("shards_glob is not specified at initialization, attempting to proceed with index_shards=True which can take a while for large objects.")
             else:
                 index_shards = False
         else:
@@ -241,7 +268,11 @@ class BaseObjectStorageDataset():
         else:            
             # query for obj sizes in order to later ignore the empty shards 
             # because they are not in the metadata dataframe
-            glob_dict = self.fs.glob(glob, detail = True)
+            if self.fs.isdir(glob):
+                glob_dict = self.fs.glob(f"{remove_trailing_forward_slash(glob)}/*", detail = True)
+            else:
+                glob_dict = self.fs.glob(glob, detail = True)
+
             kvs = glob_dict.items()
 
             objs_df = pd.DataFrame(data = map( lambda kv : (kv[1]['Key'], kv[1]['Size']), kvs ), columns = ['key', 'size'] )
@@ -262,6 +293,12 @@ class BaseObjectStorageDataset():
         self.obj_count = len(self.shards_df)
         self.row_count = self.shards_df.iloc[-1]['end_idx']
 
+    def __len__(self):
+        if self.shards_df is not None:
+            return self.shards_df.iloc[-1]['end_idx']
+        else:
+            return -1
+
     def __iter__(self):
         shard_start_idx = self.worker * self.shard_size
         while self.max_shards:
@@ -274,16 +311,24 @@ class BaseObjectStorageDataset():
                 row_end_indicies = list(row_start_indicies[1:] + [row_end_idx])
                 src_df_list = []
                 for row_start_idx, row_end_idx in zip(row_start_indicies, row_end_indicies):
+                    print(f"*** {row_start_idx=}, {row_end_idx=} ***")
 
                     #query rows based on actual indicies in the dataset
                     row_start_idx = row_start_idx % self.row_count
                     row_end_idx = row_end_idx % self.row_count
+
+                    print(f"*** {row_start_idx=}, {row_end_idx=} ***")
+
+                    print(f"*** {self.header=} ***")
+
                     if self.header is None:
                         for (skiprows, nrows, uri) in df_read_spec(self.shards_df, self.row_count, row_start_idx, row_end_idx):
                             src_df_list.append( pd.read_csv(f"filecache::{uri}", header = None, nrows=nrows, skiprows=skiprows) )
 
                     elif self.header == 'infer' or self.header == 0:                        
                         for (skiprows, nrows, uri) in df_read_spec(self.shards_df, self.row_count, row_start_idx, row_end_idx):
+                            
+                            print(f"*** {skiprows=}, {nrows=}, {uri=} ***")
 
                             if (skiprows == 0):
                                 #skiprows == 0 when we are at the top of the file with a (potential) header                            
@@ -298,10 +343,13 @@ class BaseObjectStorageDataset():
                                 # df = df.T
 
                             else:
-                                #(skiprows+1) skip the extra row  to account for the skipped header
+                                #(skiprows+1) skip the extra row to account for the skipped header
                                 df = pd.read_csv(f"filecache::{uri}", header = None, nrows = nrows, skiprows = skiprows + 1) 
 
-                            src_df_list.append(df)
+                            # empty data frames don't mess with the data
+                            # but mess with the schema so ignore them
+                            if (len(df)) > 0:
+                                src_df_list.append(df)
 
                     else:
                         raise ValueError(f"header attribute cannot be set to {self.header}")
@@ -318,4 +366,6 @@ class BaseObjectStorageDataset():
             finally:
                 shard_start_idx += (self.replicas * self.shard_size)
                 self.max_shards -= 1
-                
\ No newline at end of file
+
+def remove_trailing_forward_slash(s):
+    return re.split('/+$', s)[0]
\ No newline at end of file
